import os
import traceback
import warnings
from logging import error, info
from os import getpid
from types import TracebackType
from typing import TYPE_CHECKING, Generator, List, Optional, Type

import numpy as np
import psutil

from ...geometry import BoundingBox, Vec3Int, Vec3IntLike

if TYPE_CHECKING:
    from ..view import View


def log_memory_consumption(additional_output: str = "") -> None:
    pid = os.getpid()
    process = psutil.Process(pid)
    info(
        "Currently consuming {:.2f} GB of memory ({:.2f} GB still available) "
        "in process {}. {}".format(
            process.memory_info().rss / 1024**3,
            psutil.virtual_memory().available / 1024**3,
            pid,
            additional_output,
        )
    )


class BufferedSliceWriter:
    def __init__(
        self,
        view: "View",
        offset: Optional[Vec3IntLike] = None,
        # json_update_allowed enables the update of the bounding box and rewriting of the properties json.
        # It should be False when parallel access is intended.
        json_update_allowed: bool = True,
        # buffer_size specifies, how many slices should be aggregated until they are flushed.
        buffer_size: int = 32,
        dimension: int = 2,  # z
        *,
        relative_offset: Optional[Vec3IntLike] = None,  # in mag1
        absolute_offset: Optional[Vec3IntLike] = None,  # in mag1
        use_logging: bool = False,
    ) -> None:
        """see `View.get_buffered_slice_writer()`"""

        self.view = view
        self.buffer_size = buffer_size
        self.dtype = self.view.get_dtype()
        self.use_logging = use_logging
        self.json_update_allowed = json_update_allowed
        if offset is None and relative_offset is None and absolute_offset is None:
            relative_offset = Vec3Int.zeros()
        if offset is not None:
            warnings.warn(
                "[DEPRECATION] Using offset for a buffered slice writer is deprecated. "
                + "Please use the parameter relative_offset or absolute_offset in Mag(1) instead.",
                DeprecationWarning,
            )
        self.offset = None if offset is None else Vec3Int(offset)
        self.relative_offset = (
            None if relative_offset is None else Vec3Int(relative_offset)
        )
        self.absolute_offset = (
            None if absolute_offset is None else Vec3Int(absolute_offset)
        )
        self.dimension = dimension

        effective_offset = Vec3Int.full(0)
        if self.relative_offset is not None:
            effective_offset = self.view.bounding_box.topleft + self.relative_offset

        if self.absolute_offset is not None:
            effective_offset = self.absolute_offset

        view_chunk_depth = self.view.info.chunk_shape[self.dimension]
        if (
            effective_offset is not None
            and effective_offset[self.dimension] % view_chunk_depth != 0
        ):
            warnings.warn(
                "[WARNING] Using an offset that doesn't align with the datataset's chunk size, "
                + "will slow down the buffered slice writer, because twice as many chunks will be written.",
            )
        if buffer_size >= view_chunk_depth and buffer_size % view_chunk_depth > 0:
            warnings.warn(
                "[WARNING] Using a buffer size that doesn't align with the datataset's chunk size, "
                + "will slow down the buffered slice writer.",
            )

        assert 0 <= dimension <= 2

        self.slices_to_write: List[np.ndarray] = []
        self.current_slice: Optional[int] = None
        self.buffer_start_slice: Optional[int] = None

    def _flush_buffer(self) -> None:
        if len(self.slices_to_write) == 0:
            return

        assert (
            len(self.slices_to_write) <= self.buffer_size
        ), "The WKW buffer is larger than the defined batch_size. The buffer should have been flushed earlier. This is probably a bug in the BufferedSliceWriter."

        uniq_dtypes = set(map(lambda _slice: _slice.dtype, self.slices_to_write))
        assert (
            len(uniq_dtypes) == 1
        ), "The buffer of BufferedSliceWriter contains slices with differing dtype."
        assert uniq_dtypes.pop() == self.dtype, (
            "The buffer of BufferedSliceWriter contains slices with a dtype "
            "which differs from the dtype with which the BufferedSliceWriter was instantiated."
        )

        if self.use_logging:
            info(
                "({}) Writing {} slices at position {}.".format(
                    getpid(), len(self.slices_to_write), self.buffer_start_slice
                )
            )
            log_memory_consumption()

        try:
            assert (
                self.buffer_start_slice is not None
            ), "Failed to write buffer: The buffer_start_slice is not set."
            max_width = max(section.shape[-2] for section in self.slices_to_write)
            max_height = max(section.shape[-1] for section in self.slices_to_write)
            channel_count = self.slices_to_write[0].shape[0]

            buffer_bbox = BoundingBox(
                (0, 0, 0), (max_width, max_height, self.buffer_size)
            )

            shard_dimensions = self.view._get_file_dimensions().moveaxis(
                -1, self.dimension
            )
            chunk_size = Vec3Int(
                min(shard_dimensions[0], max_width),
                min(shard_dimensions[1], max_height),
                self.buffer_size,
            )
            for chunk_bbox in buffer_bbox.chunk(chunk_size):
                info(f"Writing chunk {chunk_bbox}")
                width, height, _ = chunk_bbox.size
                data = np.zeros(
                    (channel_count, width, height, self.buffer_size),
                    dtype=self.slices_to_write[0].dtype,
                )

                z = 0
                for section in self.slices_to_write:
                    section_chunk = section[
                        :,
                        chunk_bbox.topleft.x : chunk_bbox.bottomright.x,
                        chunk_bbox.topleft.y : chunk_bbox.bottomright.y,
                    ]
                    data[
                        :, 0 : section_chunk.shape[-2], 0 : section_chunk.shape[-1], z
                    ] = section_chunk

                    z += 1

                buffer_start = Vec3Int(
                    chunk_bbox.topleft.x, chunk_bbox.topleft.y, self.buffer_start_slice
                ).moveaxis(-1, self.dimension)
                buffer_start_mag1 = buffer_start * self.view.mag.to_vec3_int()

                data = np.moveaxis(data, -1, self.dimension + 1)

                self.view.write(
                    data,
                    offset=buffer_start.add_or_none(self.offset),
                    relative_offset=buffer_start_mag1.add_or_none(self.relative_offset),
                    absolute_offset=buffer_start_mag1.add_or_none(self.absolute_offset),
                    json_update_allowed=self.json_update_allowed,
                )
                del data

        except Exception as exc:
            error(
                "({}) An exception occurred in BufferedSliceWriter._flush_buffer with {} "
                "slices at position {}. Original error is:\n{}:{}\n\nTraceback:".format(
                    getpid(),
                    len(self.slices_to_write),
                    self.buffer_start_slice,
                    type(exc).__name__,
                    exc,
                )
            )
            traceback.print_tb(exc.__traceback__)
            error("\n")

            raise exc
        finally:
            self.slices_to_write = []

    def _get_slice_generator(self) -> Generator[None, np.ndarray, None]:
        current_slice = 0
        while True:
            data = yield  # Data gets send from the user
            if len(self.slices_to_write) == 0:
                self.buffer_start_slice = current_slice
            if len(data.shape) == 2:
                # The input data might contain channel data or not.
                # Bringing it into the same shape simplifies the code
                data = np.expand_dims(data, axis=0)
            self.slices_to_write.append(data)
            current_slice += 1

            if current_slice % self.buffer_size == 0:
                self._flush_buffer()

    def __enter__(self) -> Generator[None, np.ndarray, None]:
        gen = self._get_slice_generator()
        # It is necessary to start the generator by sending "None"
        gen.send(None)  # type: ignore
        return gen

    def __exit__(
        self,
        _type: Optional[Type[BaseException]],
        _value: Optional[BaseException],
        _tb: Optional[TracebackType],
    ) -> None:
        self._flush_buffer()



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import attr
import cattr
import numpy as np
from cattr.gen import make_dict_structure_fn, make_dict_unstructure_fn, override

from ..geometry import BoundingBox, Mag, Vec3Int
from ..utils import snake_to_camel_case, warn_deprecated
from ._array import ArrayException, BaseArray, DataFormat
from .layer_categories import LayerCategoryType


def _extract_num_channels(
    num_channels_in_properties: Optional[int],
    path: Path,
    layer: str,
    mag: Optional[Union[int, Mag]],
) -> int:
    # if a wk dataset is not created with this API, then it most likely doesn't have the attribute 'numChannels' in the
    # datasource-properties.json. In this case we need to extract the number of channels from the 'header.wkw'.
    if num_channels_in_properties is not None:
        return num_channels_in_properties

    if mag is None:
        # Unable to extract the 'num_channels' from the 'header.wkw' if the dataset has no magnifications.
        # This should never be the case because wkw-datasets that are created without this API always have a magnification.
        raise RuntimeError(
            "Cannot extract the number of channels of a dataset without a properties file and without any magnifications"
        )

    mag = Mag(mag)
    array_file_path = path / layer / mag.to_layer_name()
    try:
        array = BaseArray.open(array_file_path)
    except ArrayException as e:
        raise Exception(
            f"The dataset you are trying to open does not have the attribute 'numChannels' for layer {layer}. "
            f"However, this attribute is necessary. To mitigate this problem, it was tried to locate "
            f"the file {array_file_path} to extract the num_channels from there. "
            f"Since this file does not exist, the attempt to open the dataset failed. "
            f"Please add the attribute manually to solve the problem. "
            f"If the layer does not contain any data, you can also delete the layer and add it again.",
        ) from e
    return array.info.num_channels


_properties_floating_type_to_python_type: Dict[Union[str, type], np.dtype] = {
    "float": np.dtype("float32"),
    #  np.float: np.dtype("float32"),  # np.float is an alias for float
    float: np.dtype("float32"),
    "double": np.dtype("float64"),
}

_python_floating_type_to_properties_type = {
    "float32": "float",
    "float64": "double",
}


# --- View configuration --------------------


@attr.define
class DatasetViewConfiguration:
    """
    Stores information on how the dataset is shown in webknossos by default.
    """

    four_bit: Optional[bool] = None
    interpolation: Optional[bool] = None
    render_missing_data_black: Optional[bool] = None
    loading_strategy: Optional[str] = None
    segmentation_pattern_opacity: Optional[int] = None
    zoom: Optional[float] = None
    position: Optional[Tuple[int, int, int]] = None
    rotation: Optional[Tuple[int, int, int]] = None


@attr.define
class LayerViewConfiguration:
    """
    Stores information on how the dataset is shown in webknossos by default.
    """

    color: Optional[Tuple[int, int, int]] = None
    """Color in RGB from 0 to 255. The WEBKNOSSOS default is [255, 255, 255]."""

    alpha: Optional[float] = None
    """Alpha value from 0 to 100. The WEBKNOSSOS default is 100 except for segmentation
    layers where it is 20."""

    intensity_range: Optional[Tuple[float, float]] = None
    """Min and max data value range (dependent on the layer's data type). Can be used to threshold the value range.
    The WEBKNOSSOS default is the full value range."""

    min: Optional[float] = None  # pylint: disable=redefined-builtin
    """Minimum data value that might be encountered. This will restrict the histogram in WEBKNOSSOS and possibly overwrite
    the min value of the `intensityRange` (if that is lower)."""

    max: Optional[float] = None  # pylint: disable=redefined-builtin
    """Maximum data value that might be encountered. This will restrict the histogram in WEBKNOSSOS and possibly overwrite
    the max value of the `intensityRange` (if that is higher)."""

    is_disabled: Optional[bool] = None
    """Disable a layer. The WEBKNOSSOS default is False."""

    is_inverted: Optional[bool] = None
    """Invert a layer. The WEBKNOSSOS default is False."""

    is_in_edit_mode: Optional[bool] = None
    """Enable the histogram edit mode. The WEBKNOSSOS default is False."""


# --- Property --------------------


@attr.define
class MagViewProperties:
    mag: Mag
    cube_length: Optional[int] = None
    axis_order: Optional[Dict[str, int]] = None

    @property
    def resolution(self) -> Mag:
        warn_deprecated("resolution", "mag")
        return self.mag


@attr.define
class LayerProperties:
    name: str
    category: LayerCategoryType
    bounding_box: BoundingBox
    element_class: str
    data_format: DataFormat
    mags: List[MagViewProperties]
    num_channels: Optional[int] = None
    default_view_configuration: Optional[LayerViewConfiguration] = None

    @property
    def resolutions(self) -> List[MagViewProperties]:
        warn_deprecated("resolutions", "mags")
        return self.mags


@attr.define
class SegmentationLayerProperties(LayerProperties):
    largest_segment_id: Optional[int] = None
    mappings: List[str] = []


@attr.define
class DatasetProperties:
    id: Dict[str, str]
    scale: Tuple[float, float, float]
    data_layers: List[
        Union[
            SegmentationLayerProperties,
            LayerProperties,
        ]
    ]
    default_view_configuration: Optional[DatasetViewConfiguration] = None


# --- Converter --------------------

dataset_converter = cattr.Converter()

# register (un-)structure hooks for non-attr-classes
bbox_to_wkw: Callable[[BoundingBox], dict] = lambda o: o.to_wkw_dict()
dataset_converter.register_unstructure_hook(BoundingBox, bbox_to_wkw)
dataset_converter.register_structure_hook(
    BoundingBox, lambda d, _: BoundingBox.from_wkw_dict(d)
)


def mag_unstructure(mag: Mag) -> List[int]:
    return mag.to_list()


dataset_converter.register_unstructure_hook(Mag, mag_unstructure)
dataset_converter.register_structure_hook(Mag, lambda d, _: Mag(d))

vec3int_to_array: Callable[[Vec3Int], List[int]] = lambda o: o.to_list()
dataset_converter.register_unstructure_hook(Vec3Int, vec3int_to_array)
dataset_converter.register_structure_hook(
    Vec3Int, lambda d, _: Vec3Int.full(d) if isinstance(d, int) else Vec3Int(d)
)

dataset_converter.register_structure_hook_func(
    lambda d: d == LayerCategoryType, lambda d, _: str(d)  # type: ignore[comparison-overlap]
)

# Register (un-)structure hooks for attr-classes to bring the data into the expected format.
# The properties on disk (in datasource-properties.json) use camel case for the names of the attributes.
# However, we use snake case for the attribute names in python.
# This requires that the names of the attributes are renamed during (un-)structuring.
# Additionally we only want to unstructure attributes which don't have the default value
# (e.g. Layer.default_view_configuration has many attributes which are all optionally).
for cls in [
    DatasetProperties,
    MagViewProperties,
    DatasetViewConfiguration,
    LayerViewConfiguration,
]:
    dataset_converter.register_unstructure_hook(
        cls,
        make_dict_unstructure_fn(
            cls,
            dataset_converter,
            **{
                a.name: override(
                    omit_if_default=True, rename=snake_to_camel_case(a.name)
                )
                for a in attr.fields(cls)  # type: ignore[misc]
            },
        ),
    )
    dataset_converter.register_structure_hook(
        cls,
        make_dict_structure_fn(
            cls,
            dataset_converter,
            **{
                a.name: override(rename=snake_to_camel_case(a.name))
                for a in attr.fields(cls)  # type: ignore[misc]
            },
        ),
    )


# The serialization of `LayerProperties` differs slightly based on whether it is a `wkw` or `zarr` layer.
# These post-unstructure and pre-structure functions perform the conditional field renames.
def mag_view_properties_post_structure(d: Dict[str, Any]) -> Dict[str, Any]:
    d["resolution"] = d["mag"]
    del d["mag"]
    return d


def mag_view_properties_pre_unstructure(d: Dict[str, Any]) -> Dict[str, Any]:
    d["mag"] = d["resolution"]
    del d["resolution"]
    return d


def layer_properties_post_unstructure(
    converter_fn: Callable[
        [Union[LayerProperties, SegmentationLayerProperties]], Dict[str, Any]
    ]
) -> Callable[[Union[LayerProperties, SegmentationLayerProperties]], Dict[str, Any]]:
    def __layer_properties_post_unstructure(
        obj: Union[LayerProperties, SegmentationLayerProperties]
    ) -> Dict[str, Any]:
        d = converter_fn(obj)
        if d["dataFormat"] == "wkw":
            d["wkwResolutions"] = [
                mag_view_properties_post_structure(m) for m in d["mags"]
            ]
            del d["mags"]
        return d

    return __layer_properties_post_unstructure


def layer_properties_pre_structure(
    converter_fn: Callable[
        [Dict[str, Any], Type[Union[LayerProperties, SegmentationLayerProperties]]],
        Union[LayerProperties, SegmentationLayerProperties],
    ]
) -> Callable[
    [Any, Type[Union[LayerProperties, SegmentationLayerProperties]]],
    Union[LayerProperties, SegmentationLayerProperties],
]:
    def __layer_properties_pre_structure(
        d: Dict[str, Any],
        type_value: Type[Union[LayerProperties, SegmentationLayerProperties]],
    ) -> Union[LayerProperties, SegmentationLayerProperties]:
        if d["dataFormat"] == "wkw":
            d["mags"] = [
                mag_view_properties_pre_unstructure(m) for m in d["wkwResolutions"]
            ]
            del d["wkwResolutions"]
        obj = converter_fn(d, type_value)
        return obj

    return __layer_properties_pre_structure


for cls in [
    LayerProperties,
    SegmentationLayerProperties,
]:
    dataset_converter.register_unstructure_hook(
        cls,
        layer_properties_post_unstructure(
            make_dict_unstructure_fn(
                cls,
                dataset_converter,
                **{
                    a.name: override(
                        omit_if_default=True, rename=snake_to_camel_case(a.name)
                    )
                    for a in attr.fields(cls)  # type: ignore[misc]
                },
            )
        ),
    )
    dataset_converter.register_structure_hook(
        cls,
        layer_properties_pre_structure(
            make_dict_structure_fn(
                cls,
                dataset_converter,
                **{
                    a.name: override(rename=snake_to_camel_case(a.name))
                    for a in attr.fields(cls)  # type: ignore[misc]
                },
            )
        ),
    )


# Disambiguation of Unions only work automatically if the two attrs-classes have at least 1 unique attribute
# This is not the case here because SegmentationLayerProperties inherits LayerProperties
def disambiguate_layer_properties(obj: dict, _: Any) -> LayerProperties:
    if obj["category"] == "color":
        return dataset_converter.structure(obj, LayerProperties)
    elif obj["category"] == "segmentation":
        return dataset_converter.structure(obj, SegmentationLayerProperties)
    else:
        raise RuntimeError(
            "Failed to read the properties of a layer: the category has to be `color` or `segmentation`."
        )


dataset_converter.register_structure_hook(
    Union[
        SegmentationLayerProperties,
        LayerProperties,
    ],
    disambiguate_layer_properties,
)



import logging
from os import PathLike
from pathlib import Path
from typing import List, Optional, TypeVar, Union, cast

import numpy as np
from rich.progress import track

from ..dataset import Dataset, LayerCategoryType
from ..dataset.properties import LayerViewConfiguration, dataset_converter
from ..geometry import BoundingBox, Mag, Vec3Int
from .context import _get_context

logger = logging.getLogger(__name__)


T = TypeVar("T")


_DOWNLOAD_CHUNK_SHAPE = Vec3Int(512, 512, 512)


def download_dataset(
    dataset_name: str,
    organization_id: str,
    sharing_token: Optional[str] = None,
    bbox: Optional[BoundingBox] = None,
    layers: Optional[List[str]] = None,
    mags: Optional[List[Mag]] = None,
    path: Optional[Union[PathLike, str]] = None,
    exist_ok: bool = False,
) -> Dataset:
    context = _get_context()
    api_client = context.api_client
    api_dataset = api_client.dataset_info(
        organization_id, dataset_name, sharing_token=sharing_token
    )

    datastore_client = context.get_datastore_api_client(api_dataset.data_store.url)
    optional_datastore_token = sharing_token or context.datastore_token

    actual_path = Path(dataset_name) if path is None else Path(path)
    if actual_path.exists():
        logger.warning(f"{actual_path} already exists, skipping download.")
        return Dataset.open(actual_path)

    api_data_layers = api_dataset.data_source.data_layers
    voxel_size = api_dataset.data_source.scale
    if api_data_layers is None or len(api_data_layers) == 0 or voxel_size is None:
        raise RuntimeError(
            f"Could not download dataset {dataset_name}: {api_dataset.data_source.status or 'Unknown error.'}"
        )
    dataset = Dataset(
        actual_path, name=api_dataset.name, voxel_size=voxel_size, exist_ok=exist_ok
    )
    for layer_name in layers or [i.name for i in api_data_layers]:
        matching_api_data_layers = [i for i in api_data_layers if i.name == layer_name]
        assert (
            len(matching_api_data_layers) > 0
        ), f"The provided layer name {layer_name} could not be found in the requested dataset."
        assert (
            len(matching_api_data_layers) == 1
        ), f"The provided layer name {layer_name} was found multiple times in the requested dataset."
        api_data_layer = matching_api_data_layers[0]
        category = cast(LayerCategoryType, api_data_layer.category)
        layer = dataset.add_layer(
            layer_name=layer_name,
            category=category,
            dtype_per_layer=api_data_layer.element_class,
            num_channels=3 if api_data_layer.element_class == "uint24" else 1,
            largest_segment_id=api_data_layer.largest_segment_id,
        )

        if api_data_layer.default_view_configuration is not None:
            default_view_configuration = dataset_converter.structure(
                api_data_layer.default_view_configuration, LayerViewConfiguration
            )
            layer.default_view_configuration = default_view_configuration

        if bbox is None:
            response_bbox = api_data_layer.bounding_box
            layer.bounding_box = BoundingBox(
                Vec3Int(response_bbox.top_left),
                Vec3Int(response_bbox.width, response_bbox.height, response_bbox.depth),
            )
        else:
            assert isinstance(
                bbox, BoundingBox
            ), f"Expected a BoundingBox object for the bbox parameter but got {type(bbox)}"
            layer.bounding_box = bbox
        if mags is None:
            mags = [Mag(mag) for mag in api_data_layer.resolutions]
        for mag in mags:
            mag_view = layer.get_or_add_mag(
                mag,
                compress=True,
                chunk_shape=Vec3Int.full(32),
                chunks_per_shard=_DOWNLOAD_CHUNK_SHAPE // 32,
            )
            aligned_bbox = layer.bounding_box.align_with_mag(mag, ceil=True)
            download_chunk_shape_in_mag = _DOWNLOAD_CHUNK_SHAPE * mag.to_vec3_int()
            for chunk in track(
                list(
                    aligned_bbox.chunk(
                        download_chunk_shape_in_mag, download_chunk_shape_in_mag
                    )
                ),
                description=f"Downloading layer={layer.name} mag={mag}",
            ):
                chunk_in_mag = chunk.in_mag(mag)
                chunk_bytes, missing_buckets = datastore_client.dataset_get_raw_data(
                    organization_name=organization_id,
                    dataset_name=dataset_name,
                    data_layer_name=layer_name,
                    mag=mag.to_long_layer_name(),
                    token=optional_datastore_token,
                    x=chunk.topleft.x,
                    y=chunk.topleft.y,
                    z=chunk.topleft.z,
                    width=chunk_in_mag.size.x,
                    height=chunk_in_mag.size.y,
                    depth=chunk_in_mag.size.z,
                )
                assert (
                    missing_buckets == "[]"
                ), f"Download contained missing buckets {missing_buckets}."
                data = np.frombuffer(
                    chunk_bytes, dtype=layer.dtype_per_channel
                ).reshape(layer.num_channels, *chunk_in_mag.size, order="F")
                mag_view.write(data, absolute_offset=chunk.topleft)
    return dataset



"""This module checks equality of two different WEBKNOSSOS datasets."""

import logging
from argparse import Namespace
from multiprocessing import cpu_count
from typing import Any, Optional

import typer
from typing_extensions import Annotated

from ..dataset import Dataset, Layer
from ..utils import get_executor_for_args
from ._utils import DistributionStrategy, parse_path

logger = logging.getLogger(__name__)


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            help="Path to your first WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    target: Annotated[
        Any,
        typer.Argument(
            help="Path to your second WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    layer_name: Annotated[
        Optional[str],
        typer.Option(
            help="Name of the layer to compare (if not provided, all layers are compared)."
        ),
    ] = None,
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help="Necessary when using slurm as distribution strategy. Should be a JSON string "
            '(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Check equality of two WEBKNOSSOS datasets."""

    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )

    source_dataset = Dataset.open(source)
    target_dataset = Dataset.open(target)
    source_layer_names = set(source_dataset.layers.keys())
    target_layer_names = set(target_dataset.layers.keys())

    layer_names = list(source_layer_names)

    try:
        if layer_name is not None:
            assert (
                layer_name in source_layer_names
            ), f"Provided layer {layer_name} does not exist in source dataset."
            assert (
                layer_name in target_layer_names
            ), f"Provided layer {layer_name} does not exist in target dataset."
            layer_names = [layer_name]

        else:
            assert (
                source_layer_names == target_layer_names
            ), f"The provided input datasets have different \
    layers: {source_layer_names} != {target_layer_names}"

        for name in layer_names:
            compare_layers(
                source_dataset.get_layer(name),
                target_dataset.get_layer(name),
                executor_args,
            )

        print(
            f"The datasets {source} and {target} are equal \
    (with regard to the layers: {layer_names})"
        )
    except AssertionError as err:
        print(f"The datasets are not equal: {err}")
        exit(1)


def compare_layers(
    source_layer: Layer,
    target_layer: Layer,
    executor_args: Namespace,
) -> None:
    """Compares one layer with another layer"""

    layer_name = source_layer.name
    logging.info("Checking layer_name: %s", layer_name)

    assert (
        source_layer.bounding_box == target_layer.bounding_box
    ), f"The bounding boxes of '{layer_name}' layer of source and target \
are not equal: {source_layer.bounding_box} != {target_layer.bounding_box}"

    source_mags = set(source_layer.mags.keys())
    target_mags = set(target_layer.mags.keys())

    assert (
        source_mags == target_mags
    ), f"The mags of '{layer_name}' layer of source and target are not equal: \
{source_mags} != {target_mags}"

    for mag in source_mags:
        source_mag = source_layer.mags[mag]
        target_mag = target_layer.mags[mag]

        logging.info("Start verification of %s in mag %s", layer_name, mag)
        with get_executor_for_args(args=executor_args) as executor:
            assert source_mag.content_is_equal(
                target_mag, executor=executor
            ), f"The contents of {source_mag} and {target_mag} differ."



"""This module converts a KNOSSOS dataset into a WEBKNOSSOS dataset."""

import logging
import re
from argparse import Namespace
from collections import namedtuple
from functools import partial
from multiprocessing import cpu_count
from os import sep
from pathlib import Path
from types import TracebackType
from typing import Any, Generator, Iterator, Optional, Tuple, Type, Union, cast

import numpy as np
import typer
from typing_extensions import Annotated

from ..dataset import COLOR_CATEGORY, DataFormat, Dataset, View
from ..dataset.defaults import DEFAULT_CHUNK_SHAPE, DEFAULT_CHUNKS_PER_SHARD
from ..geometry import BoundingBox, Mag, Vec3Int
from ..utils import get_executor_for_args, time_start, time_stop
from ._utils import (
    DistributionStrategy,
    VoxelSize,
    parse_mag,
    parse_path,
    parse_vec3int,
    parse_voxel_size,
)

KNOSSOS_CUBE_EDGE_LEN = 128
KNOSSOS_CUBE_SIZE = KNOSSOS_CUBE_EDGE_LEN**3
KNOSSOS_CUBE_SHAPE = (KNOSSOS_CUBE_EDGE_LEN,) * 3
KNOSSOS_CUBE_REGEX = re.compile(
    rf"x(\d+){re.escape(sep)}y(\d+){re.escape(sep)}z(\d+){re.escape(sep)}(.*\.raw)$"
)

KnossosDatasetInfo = namedtuple("KnossosDatasetInfo", ("dataset_path", "dtype"))


class KnossosDataset:
    def __init__(self, root: Union[str, Path], dtype: np.dtype):
        self.root = Path(root)
        self.dtype = dtype

    def read(
        self, offset: Tuple[int, int, int], shape: Tuple[int, int, int]
    ) -> np.ndarray:
        assert offset[0] % KNOSSOS_CUBE_EDGE_LEN == 0
        assert offset[1] % KNOSSOS_CUBE_EDGE_LEN == 0
        assert offset[2] % KNOSSOS_CUBE_EDGE_LEN == 0
        assert shape == KNOSSOS_CUBE_SHAPE
        return self.read_cube(tuple(x // KNOSSOS_CUBE_EDGE_LEN for x in offset))

    def write(self, offset: Tuple[int, int, int], data: np.ndarray) -> None:
        assert offset[0] % KNOSSOS_CUBE_EDGE_LEN == 0
        assert offset[1] % KNOSSOS_CUBE_EDGE_LEN == 0
        assert offset[2] % KNOSSOS_CUBE_EDGE_LEN == 0
        assert data.shape == KNOSSOS_CUBE_SHAPE
        self.write_cube(tuple(x // KNOSSOS_CUBE_EDGE_LEN for x in offset), data)

    def read_cube(self, cube_xyz: Tuple[int, ...]) -> np.ndarray:
        filename = self.__get_only_raw_file_path(cube_xyz)
        if filename is None:
            return np.zeros(KNOSSOS_CUBE_SHAPE, dtype=self.dtype)
        with open(filename, "rb") as cube_file:
            cube_data = np.fromfile(cube_file, dtype=self.dtype)
            if cube_data.size != KNOSSOS_CUBE_SIZE:
                padded_data = np.zeros(KNOSSOS_CUBE_SIZE, dtype=self.dtype)
                padded_data[0 : min(cube_data.size, KNOSSOS_CUBE_SIZE)] = cube_data[
                    0 : min(cube_data.size, KNOSSOS_CUBE_SIZE)
                ]
                cube_data = padded_data
            cube_data = cube_data.reshape(KNOSSOS_CUBE_SHAPE, order="F")
            return cube_data

    def write_cube(self, cube_xyz: Tuple[int, ...], cube_data: np.ndarray) -> None:
        filename = self.__get_only_raw_file_path(cube_xyz)
        if filename is None:
            filename = self.__get_cube_folder(cube_xyz) / self.__get_cube_file_name(
                cube_xyz
            )

        filename.parent.mkdir(parents=True, exist_ok=True)
        with open(filename, "wb") as cube_file:
            cube_data.ravel(order="F").tofile(cube_file)

    def __get_cube_folder(self, cube_xyz: Tuple[int, ...]) -> Path:
        x, y, z = cube_xyz
        return (
            self.root / "x{:04d}".format(x) / "y{:04d}".format(y) / "z{:04d}".format(z)
        )

    def __get_cube_file_name(self, cube_xyz: Tuple[int, ...]) -> Path:
        x, y, z = cube_xyz
        return Path("cube_x{:04d}_y{:04d}_z{:04d}.raw".format(x, y, z))

    def __get_only_raw_file_path(self, cube_xyz: Tuple[int, ...]) -> Optional[Path]:
        cube_folder = self.__get_cube_folder(cube_xyz)
        raw_files = list(cube_folder.glob("*.raw"))
        assert len(raw_files) <= 1, "Found %d .raw files in %s" % (
            len(raw_files),
            cube_folder,
        )
        return raw_files[0] if len(raw_files) > 0 else None

    def list_files(self) -> Iterator[Path]:
        return self.root.glob("*/*/*/*.raw")

    def __parse_cube_file_name(self, filename: Path) -> Optional[Tuple[int, int, int]]:
        m = KNOSSOS_CUBE_REGEX.search(str(filename))
        if m is None:
            return None
        return int(m.group(1)), int(m.group(2)), int(m.group(3))

    def list_cubes(self) -> Generator[Tuple[int, int, int], Any, None]:
        return (
            f
            for f in (self.__parse_cube_file_name(f) for f in self.list_files())
            if f is not None
        )

    def close(self) -> None:
        pass

    @staticmethod
    def open(root: Union[str, Path], dtype: np.dtype) -> "KnossosDataset":
        return KnossosDataset(root, dtype)

    def __enter__(self) -> "KnossosDataset":
        return self

    def __exit__(
        self,
        _type: Optional[Type[BaseException]],
        _value: Optional[BaseException],
        _tb: Optional[TracebackType],
    ) -> None:
        self.close()


def open_knossos(info: KnossosDatasetInfo) -> KnossosDataset:
    return KnossosDataset.open(info.dataset_path, np.dtype(info.dtype))


def convert_cube_job(
    source_knossos_info: KnossosDatasetInfo, args: Tuple[View, int]
) -> None:
    target_view, _ = args

    time_start(f"Converting of {target_view.bounding_box}")
    cube_size = cast(Tuple[int, int, int], (KNOSSOS_CUBE_EDGE_LEN,) * 3)

    offset = target_view.bounding_box.in_mag(target_view.mag).topleft
    size = target_view.bounding_box.in_mag(target_view.mag).size
    buffer = np.zeros(size.to_tuple(), dtype=target_view.get_dtype())
    with open_knossos(source_knossos_info) as source_knossos:
        for x in range(0, size.x, KNOSSOS_CUBE_EDGE_LEN):
            for y in range(0, size.y, KNOSSOS_CUBE_EDGE_LEN):
                for z in range(0, size.z, KNOSSOS_CUBE_EDGE_LEN):
                    cube_data = source_knossos.read(
                        (offset + Vec3Int(x, y, z)).to_tuple(), cube_size
                    )
                    buffer[
                        x : (x + KNOSSOS_CUBE_EDGE_LEN),
                        y : (y + KNOSSOS_CUBE_EDGE_LEN),
                        z : (z + KNOSSOS_CUBE_EDGE_LEN),
                    ] = cube_data
    target_view.write(buffer)

    time_stop(f"Converting of {target_view.bounding_box}")


def convert_knossos(
    source_path: Path,
    target_path: Path,
    layer_name: str,
    dtype: str,
    voxel_size: Tuple[float, float, float],
    data_format: DataFormat,
    chunk_shape: Vec3Int,  # in target-mag
    chunks_per_shard: Vec3Int,
    mag: Mag = Mag(1),
    args: Optional[Namespace] = None,
) -> None:
    """Performs the conversion of a KNOSSOS dataset to a WEBKNOSSOS dataset."""

    source_knossos_info = KnossosDatasetInfo(source_path, dtype)

    target_dataset = Dataset(target_path, voxel_size, exist_ok=True)
    target_layer = target_dataset.get_or_add_layer(
        layer_name,
        COLOR_CATEGORY,
        data_format=data_format,
        dtype_per_channel=dtype,
    )

    with open_knossos(source_knossos_info) as source_knossos:
        knossos_cubes = np.array(list(source_knossos.list_cubes()))
        if len(knossos_cubes) == 0:
            logging.error(
                "No input KNOSSOS cubes found. Make sure to pass the path which "
                "points to a KNOSSOS magnification (e.g., testdata/knossos/color/1)."
            )
            exit(1)

        min_xyz = knossos_cubes.min(axis=0) * KNOSSOS_CUBE_EDGE_LEN
        max_xyz = (knossos_cubes.max(axis=0) + 1) * KNOSSOS_CUBE_EDGE_LEN
        target_layer.bounding_box = BoundingBox(
            Vec3Int(min_xyz), Vec3Int(max_xyz - min_xyz)
        )

    target_mag = target_layer.get_or_add_mag(
        mag, chunk_shape=chunk_shape, chunks_per_shard=chunks_per_shard
    )

    with get_executor_for_args(args) as executor:
        target_mag.for_each_chunk(
            partial(convert_cube_job, source_knossos_info),
            chunk_shape=chunk_shape * mag * chunks_per_shard,
            executor=executor,
            progress_desc=f"Converting knossos layer {layer_name}",
        )


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            help="Path to your image data.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    target: Annotated[
        Any,
        typer.Argument(
            help="Target path to save your WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    layer_name: Annotated[
        str,
        typer.Option(help="Name of the cubed layer (color or segmentation)"),
    ] = "color",
    voxel_size: Annotated[
        VoxelSize,
        typer.Option(
            help="The size of one voxel in source data in nanometers. "
            "Should be a comma seperated string (e.g. 11.0,11.0,20.0).",
            parser=parse_voxel_size,
            metavar="VOXEL_SIZE",
            show_default=False,
        ),
    ],
    dtype: Annotated[
        str, typer.Option(help="Target datatype (e.g. uint8, uint16, uint32)")
    ] = "uint8",
    data_format: Annotated[
        DataFormat,
        typer.Option(
            help="Data format to store the target dataset in.",
        ),
    ] = "wkw",  # type:ignore
    chunk_shape: Annotated[
        Vec3Int,
        typer.Option(
            help="Number of voxels to be stored as a chunk in the output format "
            "(e.g. `32` or `32,32,32`).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = DEFAULT_CHUNK_SHAPE,
    chunks_per_shard: Annotated[
        Vec3Int,
        typer.Option(
            help="Number of chunks to be stored as a shard in the output format "
            "(e.g. `32` or `32,32,32`).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = DEFAULT_CHUNKS_PER_SHARD,
    mag: Annotated[
        Mag,
        typer.Option(
            help="Mag to start upsampling from. "
            "Should be number or minus seperated string (e.g. 2 or 2-2-2).",
            parser=parse_mag,
            metavar="MAG",
        ),
    ] = 1,  # type: ignore
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help="Necessary when using slurm as distribution strategy. Should be a JSON string "
            '(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Convert your KNOSSOS dataset to a WEBKNOSOOS dataset."""

    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )

    convert_knossos(
        source,
        target,
        layer_name,
        dtype,
        voxel_size,
        data_format,
        chunk_shape,
        chunks_per_shard,
        mag,
        executor_args,
    )



import re
from operator import add, floordiv, mod, mul, sub
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union, cast

import numpy as np

value_error = "Vector components must be three integers or a Vec3IntLike object."


class Vec3Int(tuple):
    def __new__(
        cls,
        vec: Union[int, "Vec3IntLike"],
        y: Optional[int] = None,
        z: Optional[int] = None,
    ) -> "Vec3Int":
        """
        Class to represent a 3D vector. Inherits from tuple and provides useful
        methods and operations on top.

        A small usage example:

        ```python
        from webknossos import Vec3Int

        vector_1 = Vec3Int(1, 2, 3)
        vector_2 = Vec3Int.full(1)
        assert vector_2.x == vector_2.y == vector_2.y

        assert vector_1 + vector_2 == (2, 3, 4)
        ```
        """

        if isinstance(vec, Vec3Int):
            return vec

        as_tuple: Optional[Tuple[int, int, int]] = None

        if isinstance(vec, int):
            assert y is not None and z is not None, value_error
            assert isinstance(y, int) and isinstance(z, int), value_error
            as_tuple = vec, y, z
        else:
            assert y is None and z is None, value_error
            if isinstance(vec, np.ndarray):
                assert np.count_nonzero(vec % 1) == 0, value_error
                assert vec.shape == (
                    3,
                ), "Numpy array for Vec3Int must have shape (3,)."
            if isinstance(vec, Iterable):
                as_tuple = cast(Tuple[int, int, int], tuple(int(item) for item in vec))
                assert len(as_tuple) == 3, value_error
        assert as_tuple is not None and len(as_tuple) == 3, value_error

        return super().__new__(cls, cast(Iterable, as_tuple))

    @staticmethod
    def from_xyz(x: int, y: int, z: int) -> "Vec3Int":
        """Use Vec3Int.from_xyz for fast construction."""

        # By calling __new__ of tuple directly, we circumvent
        # the tolerant (and potentially) slow Vec3Int.__new__ method.
        return tuple.__new__(Vec3Int, (x, y, z))

    @staticmethod
    def from_vec3_float(vec: Tuple[float, float, float]) -> "Vec3Int":
        return Vec3Int(int(vec[0]), int(vec[1]), int(vec[2]))

    @staticmethod
    def from_vec_or_int(vec_or_int: Union["Vec3IntLike", int]) -> "Vec3Int":
        if isinstance(vec_or_int, int):
            return Vec3Int.full(vec_or_int)
        else:
            return Vec3Int(vec_or_int)

    @staticmethod
    def from_str(string: str) -> "Vec3Int":
        if re.match(r"\(\d+,\d+,\d+\)", string):
            return Vec3Int(tuple(map(int, re.findall(r"\d+", string))))
        else:
            return Vec3Int.full(int(string))

    @property
    def x(self) -> int:
        return self[0]

    @property
    def y(self) -> int:
        return self[1]

    @property
    def z(self) -> int:
        return self[2]

    def with_x(self, new_x: int) -> "Vec3Int":
        return Vec3Int.from_xyz(new_x, self.y, self.z)

    def with_y(self, new_y: int) -> "Vec3Int":
        return Vec3Int.from_xyz(self.x, new_y, self.z)

    def with_z(self, new_z: int) -> "Vec3Int":
        return Vec3Int.from_xyz(self.x, self.y, new_z)

    def to_np(self) -> np.ndarray:
        return np.array((self.x, self.y, self.z))

    def to_list(self) -> List[int]:
        return [self.x, self.y, self.z]

    def to_tuple(self) -> Tuple[int, int, int]:
        return self.x, self.y, self.z

    def contains(self, needle: int) -> bool:
        return self.x == needle or self.y == needle or self.z == needle

    def is_positive(self, strictly_positive: bool = False) -> bool:
        if strictly_positive:
            return all(i > 0 for i in self)
        else:
            return all(i >= 0 for i in self)

    def is_uniform(self) -> bool:
        return self.x == self.y == self.z

    def _element_wise(
        self, other: Union[int, "Vec3IntLike"], fn: Callable[[int, Any], int]
    ) -> "Vec3Int":
        if isinstance(other, int):
            other_imported = Vec3Int.from_xyz(other, other, other)
        else:
            other_imported = Vec3Int(other)
        return Vec3Int.from_xyz(
            fn(self.x, other_imported.x),
            fn(self.y, other_imported.y),
            fn(self.z, other_imported.z),
        )

    # note: (arguments incompatible with superclass, do not add Vec3Int to plain tuple! Hence the type:ignore)
    def __add__(self, other: Union[int, "Vec3IntLike"]) -> "Vec3Int":  # type: ignore[override]
        return self._element_wise(other, add)

    def __sub__(self, other: Union[int, "Vec3IntLike"]) -> "Vec3Int":
        return self._element_wise(other, sub)

    # Note: When multiplying regular tuples with an int those are repeated,
    # which is a different behavior in the superclass! Hence the type:ignore.
    def __mul__(self, other: Union[int, "Vec3IntLike"]) -> "Vec3Int":  # type: ignore[override]
        return self._element_wise(other, mul)

    def __floordiv__(self, other: Union[int, "Vec3IntLike"]) -> "Vec3Int":
        return self._element_wise(other, floordiv)

    def __mod__(self, other: Union[int, "Vec3IntLike"]) -> "Vec3Int":
        return self._element_wise(other, mod)

    def __neg__(self) -> "Vec3Int":
        return Vec3Int.from_xyz(-self.x, -self.y, -self.z)

    def ceildiv(self, other: Union[int, "Vec3IntLike"]) -> "Vec3Int":
        return (self + other - 1) // other

    def pairmax(self, other: Union[int, "Vec3IntLike"]) -> "Vec3Int":
        return self._element_wise(other, max)

    def pairmin(self, other: Union[int, "Vec3IntLike"]) -> "Vec3Int":
        return self._element_wise(other, min)

    def prod(self) -> int:
        return self.x * self.y * self.z

    def __repr__(self) -> str:
        return f"Vec3Int({self.x},{self.y},{self.z})"

    def add_or_none(self, other: Optional["Vec3Int"]) -> Optional["Vec3Int"]:
        return None if other is None else self + other

    def moveaxis(
        self, source: Union[int, List[int]], target: Union[int, List[int]]
    ) -> "Vec3Int":
        """
        Allows to move one element at index `source` to another index `target`. Similar to
        np.moveaxis, this is *not* a swap operation but instead it moves the specified
        source so that the other elements move when necessary.
        """

        # Piggy-back on np.moveaxis by creating an auxiliary array where the indices 0, 1 and
        # 2 appear in the shape.
        indices = np.moveaxis(np.zeros((0, 1, 2)), source, target).shape
        arr = self.to_np()[np.array(indices)]
        return Vec3Int(arr)

    @classmethod
    def zeros(cls) -> "Vec3Int":
        return cls(0, 0, 0)

    @classmethod
    def ones(cls) -> "Vec3Int":
        return cls(1, 1, 1)

    @classmethod
    def full(cls, an_int: int) -> "Vec3Int":
        return cls(an_int, an_int, an_int)


Vec3IntLike = Union[Vec3Int, Tuple[int, int, int], np.ndarray, Iterable[int]]



from functools import reduce

from ...geometry import BoundingBox
from ..mag_view import MagView


def infer_bounding_box_existing_files(mag_view: MagView) -> BoundingBox:
    """Since volume annotation layers are only a single layer, they do not contain a datasource-properties.json.
    Therefore, the bounding box needs to be inferred when working with those.
    The returned bounding box is measured in Mag(1) voxels."""

    return reduce(
        lambda acc, bbox: acc.extended_by(bbox),
        mag_view.get_bounding_boxes_on_disk(),
        BoundingBox.empty(),
    )



from typing import Dict, Optional, Tuple, TypeVar

Vector3 = Tuple[float, float, float]
Vector4 = Tuple[float, float, float, float]


def filter_none_values(_dict: Dict[str, Optional[str]]) -> Dict[str, str]:
    """XML values must not be None."""
    return {key: value for key, value in _dict.items() if value is not None}


T = TypeVar("T")


def enforce_not_null(val: Optional[T]) -> T:
    if val is None:
        raise ValueError("Value is None")
    return val


def as_int_unless_none(val: Optional[str]) -> Optional[int]:
    if val is None:
        return None
    return int(val)



from typing import TYPE_CHECKING, Optional, Sequence, TypeVar, Union

from ..utils import LazyReadOnlyDict

if TYPE_CHECKING:
    from .dataset import RemoteDataset


K = TypeVar("K")  # key
V = TypeVar("V")  # value
C = TypeVar("C")  # cache


class RemoteDatasetRegistry(LazyReadOnlyDict[str, "RemoteDataset"]):
    """Dict-like class mapping dataset names to `RemoteDataset` instances."""

    def __init__(
        self,
        organization_id: Optional[str],
        tags: Optional[Union[str, Sequence[str]]],
    ) -> None:
        from ..administration.user import User
        from ..client.context import _get_api_client
        from .dataset import Dataset

        client = _get_api_client(enforce_auth=True)

        if organization_id is None:
            organization_id = User.get_current_user().organization_id

        if isinstance(tags, str):
            tags = [tags]

        dataset_infos = client.dataset_list(
            is_active=True, organization_name=organization_id
        )

        datasets_names = []

        for dataset_info in dataset_infos:
            tags_match = tags is None or any(tag in tags for tag in dataset_info.tags)
            if tags_match:
                datasets_names.append(dataset_info.name)

        super().__init__(
            entries=dict(zip(datasets_names, datasets_names)),
            func=lambda name: Dataset.open_remote(name, organization_id),
        )



"""
This package package provides methods to modify or interact with WEBKNOSSOS resources.

The core classes and can be created and saved as shown:

<div markdown="1" style="overflow-x: auto;">
<div markdown="1" style="width: max-content;">
|                  | Create ✨                                                                                                                                             | Open/Load 📂                                                                                                                                           | Save 💾                                                                       | Download 📥                                                                                       | Upload 📤                                                                         |
|------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------|--------------------------------------------------------------------------------------------------------------------------------------------------------|-------------------------------------------------------------------------------|---------------------------------------------------------------------------------------------------|------------------------------------------------------------------------------------|
| **`Dataset`**    | [`Dataset(…)`](webknossos/dataset/dataset.html#Dataset.__init__),<br/>[`Dataset.from_images(…)`](webknossos/dataset/dataset.html#Dataset.from_images) | [`Dataset.open(…)`](webknossos/dataset/dataset.html#Dataset.open),<br/>[`Dataset.open_remote(…)`](webknossos/dataset/dataset.html#Dataset.open_remote) | *N/A*[^ds-save]                                                               | [`Dataset.download(…)`](webknossos/dataset/dataset.html#Dataset.download)                         | [`dataset.upload(…)`](webknossos/dataset/dataset.html#Dataset.upload)              |
| **`Annotation`** | [`Annotation(…)`](webknossos/annotation/annotation.html#Annotation.__init__)                                                                          | [`Annotation.load(…)`](webknossos/annotation/annotation.html#Annotation.load)                                                                          | [`annotation.save(…)`](webknossos/annotation/annotation.html#Annotation.save) | [`Annotation.download(…)`](webknossos/annotation/annotation.html#Annotation.download)[^anno-down] | [`annotation.upload(…)`](webknossos/annotation/annotation.html#Annotation.upload)  |
| **`Skeleton`**   | [`Skeleton(…)`](webknossos/skeleton/skeleton.html#Skeleton.__init__)                                                                                  | [`Skeleton.load(…)`](webknossos/skeleton/skeleton.html#Skeleton.load)                                                                                  | [`skeleton.save(…)`](webknossos/skeleton/skeleton.html#Skeleton.save)         | *only in an annotation*                                                                           | *only in an annotation*                                                            |
</div>
</div>

[^ds-save]: Since the state of a dataset is continously persisted it cannot be saved.
[^anno-down]: Downloaded annotations are not persisted on disk automatically.

Additionally, we provide the geometrical primitives `Vec3Int`, `BoundingBox` and `Mag`.

The `User`, `Project` and `Task` classes provide WEBKNOSSOS server interactions for administration purposes.
Server interactions may require [authentication](webknossos/client/context.html) e.g. via `webknossos_context`.
"""

# The table above contains zero-width spaces in the code examples after each dot to enforce correct line-breaks.

from .administration import *
from .annotation import *
from .client import *
from .dataset import *
from .geometry import *
from .skeleton import *



from typing import NamedTuple, Optional
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from ..geometry import Vec3Int
from .utils import Vector4, enforce_not_null, filter_none_values


class Segment(NamedTuple):
    id: int
    name: Optional[str]
    anchor_position: Optional[Vec3Int]
    color: Optional[Vector4]

    def _dump(self, xf: XmlWriter) -> None:
        if self.anchor_position is None:
            anchor_position_dict = {}
        else:
            anchor_position_dict = {
                "anchorPositionX": str(self.anchor_position.x),
                "anchorPositionY": str(self.anchor_position.y),
                "anchorPositionZ": str(self.anchor_position.z),
            }
        if self.color is None:
            color_dict = {}
        else:
            color_dict = {
                "color.r": str(self.color[0]),
                "color.g": str(self.color[1]),
                "color.b": str(self.color[2]),
                "color.a": str(self.color[3]),
            }
        xf.tag(
            "segment",
            filter_none_values(
                {
                    "id": str(self.id),
                    "name": self.name,
                    **anchor_position_dict,
                    **color_dict,
                }
            ),
        )

    @classmethod
    def _parse(cls, nml_segment: Element) -> "Segment":
        if nml_segment.get("color.r") is None:
            color = None
        else:
            color = (
                float(enforce_not_null(nml_segment.get("color.r"))),
                float(enforce_not_null(nml_segment.get("color.g"))),
                float(enforce_not_null(nml_segment.get("color.b"))),
                float(enforce_not_null(nml_segment.get("color.a"))),
            )
        if nml_segment.get("anchorPositionX") is None:
            anchor_position = None
        else:
            anchor_position = Vec3Int(
                int(enforce_not_null(nml_segment.get("anchorPositionX"))),
                int(enforce_not_null(nml_segment.get("anchorPositionY"))),
                int(enforce_not_null(nml_segment.get("anchorPositionZ"))),
            )
        return cls(
            int(enforce_not_null(nml_segment.get("id"))),
            name=nml_segment.get("name"),
            anchor_position=anchor_position,
            color=color,
        )



from typing import List, NamedTuple, Optional
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from .segment import Segment
from .utils import enforce_not_null, filter_none_values


class Volume(NamedTuple):
    id: int
    location: Optional[
        str
    ]  # path to a ZIP file containing a wK volume annotation, may be omitted when using skip_volume_data
    # name of an already existing wK volume annotation segmentation layer:
    fallback_layer: Optional[str]
    # older wk versions did not serialize the name which is why the name is optional:
    name: Optional[str]
    segments: List[Segment]
    format: Optional[str] = None
    largest_segment_id: Optional[int] = None

    def _dump(self, xf: XmlWriter) -> None:
        xf.startTag(
            "volume",
            filter_none_values(
                {
                    "id": str(self.id),
                    "location": self.location,
                    "fallbackLayer": self.fallback_layer,
                    "name": self.name,
                    "format": self.format,
                    "largestSegmentId": str(self.largest_segment_id)
                    if self.largest_segment_id
                    else None,
                }
            ),
        )
        if self.segments is not None:
            xf.startTag("segments")
            for segment in self.segments:
                segment._dump(xf)
            xf.endTag()  # segments
        xf.endTag()  # volume

    @classmethod
    def _parse(cls, nml_volume: Element) -> "Volume":
        largest_segment_id_str = nml_volume.get("largestSegmentId", default=None)
        return cls(
            id=int(enforce_not_null(nml_volume.get("id"))),
            location=nml_volume.get("location"),
            fallback_layer=nml_volume.get("fallbackLayer", default=None),
            name=nml_volume.get("name", default=None),
            format=nml_volume.get("format", default=None),
            segments=[],
            largest_segment_id=None
            if largest_segment_id_str is None
            else int(largest_segment_id_str),
        )



import httpx


class ApiClientError(Exception):
    def message_for_response_body(self, response: httpx.Response) -> str:
        response_limit_chars = 2000

        response_str = response.content.decode("utf-8")
        shortened_label = (
            f" (showing first {response_limit_chars} of {len(response_str)} characters)"
            if (len(response_str) > response_limit_chars)
            else ""
        )
        return f"Got response status {response.status_code} with body{shortened_label}: {response_str[0:response_limit_chars]}"

    check_version_hint = """If this is unexpected, please double-check your WEBKNOSSOS URL and credentials.
If the error persists, it might be caused by a version mismatch of the python client and the WEBKNOSSOS server API version.
See https://github.com/scalableminds/webknossos-libs/releases for current releases."""

    def request_label(self, response: httpx.Response) -> str:
        if response.request is None:
            return "a WEBKNOSSOS API request"
        return f"a {response.request.method} request for URL {response.request.url}"


class UnexpectedStatusError(ApiClientError):
    def __init__(self, response: httpx.Response):
        msg = f"""An error occurred while performing {self.request_label(response)}.
{self.check_version_hint}
{self.message_for_response_body(response)}
"""
        super().__init__(msg)


class CannotHandleResponseError(ApiClientError):
    def __init__(self, response: httpx.Response):
        msg = f"""An error occurred while processing the response to {self.request_label(response)}.
{self.check_version_hint}
{self.message_for_response_body(response)}
"""

        super().__init__(msg)



import logging
import math
import warnings
from enum import Enum
from itertools import product
from typing import TYPE_CHECKING, Callable, List, Optional, Tuple

import numpy as np
from scipy.ndimage import zoom

if TYPE_CHECKING:
    from .dataset import Dataset, Layer

from ..geometry import BoundingBox, Mag, Vec3Int, Vec3IntLike
from ._array import ArrayInfo
from .layer_categories import LayerCategoryType
from .view import View


class InterpolationModes(Enum):
    MEDIAN = 0
    MODE = 1
    NEAREST = 2
    BILINEAR = 3
    BICUBIC = 4
    MAX = 5
    MIN = 6


DEFAULT_BUFFER_SHAPE = Vec3Int.full(256)


def determine_buffer_shape(array_info: ArrayInfo) -> Vec3Int:
    return min(DEFAULT_BUFFER_SHAPE, array_info.shard_shape)


def calculate_mags_to_downsample(
    from_mag: Mag,
    coarsest_mag: Mag,
    dataset_to_align_with: Optional["Dataset"],
    voxel_size: Optional[Tuple[float, float, float]],
) -> List[Mag]:
    assert np.all(from_mag.to_np() <= coarsest_mag.to_np())
    mags = []
    current_mag = from_mag
    if dataset_to_align_with is None:
        mags_to_align_with = set()
    else:
        mags_to_align_with = set(
            mag
            for layer in dataset_to_align_with.layers.values()
            for mag in layer.mags.keys()
        )
    mags_to_align_with_by_max_dim = {mag.max_dim: mag for mag in mags_to_align_with}
    assert len(mags_to_align_with) == len(
        mags_to_align_with_by_max_dim
    ), "Some layers contain different values for the same mag, this is not allowed."
    while current_mag < coarsest_mag:
        if current_mag.max_dim * 2 in mags_to_align_with_by_max_dim:
            current_mag = mags_to_align_with_by_max_dim[current_mag.max_dim * 2]
            if current_mag > coarsest_mag:
                warnings.warn(
                    "[INFO] The mag taken from another layer is larger in some dimensions than `coarsest_mag`."
                )
        elif voxel_size is None:
            # In case the sampling mode is CONSTANT_Z or ISOTROPIC:
            current_mag = Mag(np.minimum(current_mag.to_np() * 2, coarsest_mag.to_np()))
        else:
            # In case the sampling mode is ANISOTROPIC:
            current_size = current_mag.to_np() * np.array(voxel_size)
            min_value = np.min(current_size)
            min_value_bitmask = np.array(current_size == min_value)
            factor = min_value_bitmask + 1

            # Calculate the two potential magnifications.
            # Either, double all components or only double the smallest component.
            all_voxel_sized = current_size * 2
            min_voxel_sized = (
                current_size * factor
            )  # only multiply the smallest dimension

            all_voxel_sized_ratio = np.max(all_voxel_sized) / np.min(all_voxel_sized)
            min_voxel_sized_ratio = np.max(min_voxel_sized) / np.min(min_voxel_sized)

            # The smaller the ratio between the smallest dimension and the largest dimension, the better.
            if all_voxel_sized_ratio < min_voxel_sized_ratio:
                # Multiply all dimensions with "2"
                new_mag = Mag(np.minimum(current_mag.to_np() * 2, coarsest_mag.to_np()))
            else:
                # Multiply only the minimal dimension by "2".
                new_mag = Mag(
                    np.minimum(current_mag.to_np() * factor, coarsest_mag.to_np())
                )
                # In case of isotropic resolution but anisotropic mag we need to ensure unique max dims.
                # current mag: 4-4-2, voxel_size: 1,1,1 -> new_mag: 4-4-4, therefore we skip this entry.
                if new_mag.max_dim == current_mag.max_dim and new_mag != current_mag:
                    current_mag = new_mag
                    continue
            if new_mag == current_mag:
                raise RuntimeError(
                    f"The coarsest mag {coarsest_mag} can not be reached from {current_mag} with voxel_size {voxel_size}!"
                )
            current_mag = new_mag

        mags += [current_mag]

    return mags


def calculate_mags_to_upsample(
    from_mag: Mag,
    finest_mag: Mag,
    dataset_to_align_with: Optional["Dataset"],
    voxel_size: Optional[Tuple[float, float, float]],
) -> List[Mag]:
    return list(
        reversed(
            calculate_mags_to_downsample(
                finest_mag, from_mag, dataset_to_align_with, voxel_size
            )
        )
    )[1:] + [finest_mag]


def calculate_default_coarsest_mag(dataset_size: Vec3IntLike) -> Mag:
    dataset_size = Vec3Int(dataset_size)
    # The coarsest mag should have a size of ~ 100vx**2 per slice
    coarsest_x_y = max(dataset_size[0], dataset_size[1])
    # highest power of 2 larger (or equal) than coarsest_x_y divided by 100
    # The calculated factor will be used for x, y and z here. If anisotropic downsampling takes place,
    # the dimensions can still be downsampled independently according to the voxel_size.
    return Mag(max(2 ** math.ceil(math.log(coarsest_x_y / 100, 2)), 4))  # at least 4


def parse_interpolation_mode(
    interpolation_mode: str, layer_category: LayerCategoryType
) -> InterpolationModes:
    if interpolation_mode.upper() == "DEFAULT":
        return (
            InterpolationModes.MODE
            if layer_category == "segmentation"
            else InterpolationModes.MEDIAN
        )
    else:
        return InterpolationModes[interpolation_mode.upper()]


def linear_filter_3d(data: np.ndarray, factors: List[int], order: int) -> np.ndarray:
    factors_np = np.array(factors)

    return zoom(
        data,
        1 / factors_np,
        output=data.dtype,
        # 0: nearest
        # 1: bilinear
        # 2: bicubic
        order=order,
        # this does not mean nearest interpolation,
        # it corresponds to how the borders are treated.
        mode="nearest",
        prefilter=True,
    )


def non_linear_filter_3d(
    data: np.ndarray, factors: List[int], func: Callable[[np.ndarray], np.ndarray]
) -> np.ndarray:
    ds = data.shape
    assert not any((d % factor > 0 for (d, factor) in zip(ds, factors)))
    data = data.reshape((ds[0], factors[1], ds[1] // factors[1], ds[2]), order="F")
    data = data.swapaxes(0, 1)
    data = data.reshape(
        (
            factors[0] * factors[1],
            ds[0] * ds[1] // (factors[0] * factors[1]),
            factors[2],
            ds[2] // factors[2],
        ),
        order="F",
    )
    data = data.swapaxes(2, 1)
    data = data.reshape(
        (
            factors[0] * factors[1] * factors[2],
            (ds[0] * ds[1] * ds[2]) // (factors[0] * factors[1] * factors[2]),
        ),
        order="F",
    )
    data = func(data)
    data = data.reshape(
        (ds[0] // factors[0], ds[1] // factors[1], ds[2] // factors[2]), order="F"
    )
    return data


def _max(x: np.ndarray) -> np.ndarray:
    return np.max(x, axis=0)


def _min(x: np.ndarray) -> np.ndarray:
    return np.min(x, axis=0)


def _median(x: np.ndarray) -> np.ndarray:
    return np.median(x, axis=0).astype(x.dtype)


def _mode(x: np.ndarray) -> np.ndarray:
    """
    Fast mode implementation from: https://stackoverflow.com/a/35674754
    """
    # Check inputs
    ndim = x.ndim
    axis = 0
    # Sort array
    sort = np.sort(x, axis=axis)
    # Create array to transpose along the axis and get padding shape
    transpose = np.roll(np.arange(ndim)[::-1], axis)
    shape = list(sort.shape)
    shape[axis] = 1
    # Create a boolean array along strides of unique values
    strides = (
        np.concatenate(
            [
                np.zeros(shape=shape, dtype="bool"),
                np.diff(sort, axis=axis) == 0,
                np.zeros(shape=shape, dtype="bool"),
            ],
            axis=axis,
        )
        .transpose(transpose)
        .ravel()
    )
    # Count the stride lengths
    counts = np.cumsum(strides)
    counts[~strides] = np.concatenate([[0], np.diff(counts[~strides])])
    counts[strides] = 0
    # Get shape of padded counts and slice to return to the original shape
    shape_array = np.array(sort.shape)
    shape_array[axis] += 1
    shape_array = shape_array[transpose]
    slices = [slice(None)] * ndim
    slices[axis] = slice(1, None)
    # Reshape and compute final counts
    counts = counts.reshape(shape_array).transpose(transpose)[tuple(slices)] + 1

    # Find maximum counts and return modals/counts
    slices = [slice(None, i) for i in sort.shape]
    del slices[axis]
    index = np.ogrid[slices]
    index.insert(axis, np.argmax(counts, axis=axis))
    return sort[tuple(index)]


def downsample_unpadded_data(
    buffer: np.ndarray, target_mag: Mag, interpolation_mode: InterpolationModes
) -> np.ndarray:
    target_mag_np = np.array(target_mag.to_list())
    current_dimension_size = np.array(buffer.shape[1:])
    padding_size_for_downsampling = (
        target_mag_np - (current_dimension_size % target_mag_np) % target_mag_np
    )
    padding_size_for_downsampling = list(zip([0, 0, 0], padding_size_for_downsampling))
    buffer = np.pad(
        buffer, pad_width=[(0, 0)] + padding_size_for_downsampling, mode="constant"
    )
    dimension_decrease = np.array([1] + target_mag.to_list())
    downsampled_buffer_shape = np.array(buffer.shape) // dimension_decrease
    downsampled_buffer = np.empty(dtype=buffer.dtype, shape=downsampled_buffer_shape)
    for channel in range(buffer.shape[0]):
        downsampled_buffer[channel] = downsample_cube(
            buffer[channel], target_mag.to_list(), interpolation_mode
        )
    return downsampled_buffer


def downsample_cube(
    cube_buffer: np.ndarray, factors: List[int], interpolation_mode: InterpolationModes
) -> np.ndarray:
    if interpolation_mode == InterpolationModes.MODE:
        return non_linear_filter_3d(cube_buffer, factors, _mode)
    elif interpolation_mode == InterpolationModes.MEDIAN:
        return non_linear_filter_3d(cube_buffer, factors, _median)
    elif interpolation_mode == InterpolationModes.NEAREST:
        return linear_filter_3d(cube_buffer, factors, 0)
    elif interpolation_mode == InterpolationModes.BILINEAR:
        return linear_filter_3d(cube_buffer, factors, 1)
    elif interpolation_mode == InterpolationModes.BICUBIC:
        return linear_filter_3d(cube_buffer, factors, 2)
    elif interpolation_mode == InterpolationModes.MAX:
        return non_linear_filter_3d(cube_buffer, factors, _max)
    elif interpolation_mode == InterpolationModes.MIN:
        return non_linear_filter_3d(cube_buffer, factors, _min)
    else:
        raise Exception("Invalid interpolation mode: {}".format(interpolation_mode))


def downsample_cube_job(
    args: Tuple[View, View, int],
    mag_factors: Vec3Int,
    interpolation_mode: InterpolationModes,
    buffer_shape: Vec3Int,
) -> None:
    (source_view, target_view, _i) = args

    try:
        num_channels = target_view.info.num_channels
        shape = (num_channels,) + target_view.bounding_box.in_mag(
            target_view.mag
        ).size.to_tuple()
        file_buffer = np.zeros(shape, target_view.get_dtype())

        tiles = product(
            *(
                list(range(0, math.ceil(len / buffer_edge_len)))
                for len, buffer_edge_len in zip(shape[-3:], buffer_shape)
            )
        )

        for tile in tiles:
            target_offset = Vec3Int(tile) * buffer_shape
            source_offset = mag_factors * target_offset
            source_size = source_view.bounding_box.in_mag(source_view.mag).size
            source_size = (mag_factors * buffer_shape).pairmin(
                source_size - source_offset
            )

            bbox = BoundingBox(source_offset, source_size)

            cube_buffer_channels = source_view.read(
                relative_bounding_box=bbox.from_mag_to_mag1(source_view.mag),
            )

            for channel_index in range(num_channels):
                cube_buffer = cube_buffer_channels[channel_index]

                if not np.all(cube_buffer == 0):
                    # Downsample the buffer
                    data_cube = downsample_cube(
                        cube_buffer,
                        mag_factors.to_list(),
                        interpolation_mode,
                    )

                    buffer_bbox = BoundingBox(target_offset, data_cube.shape)
                    file_buffer[(channel_index,) + buffer_bbox.to_slices()] = data_cube

        # Write the downsampled buffer to target
        if source_view.info.num_channels == 1:
            file_buffer = file_buffer[0]  # remove channel dimension
        target_view.write(file_buffer)

    except Exception as exc:
        logging.error(
            f"Downsampling of target {target_view.bounding_box} failed with {exc}"
        )
        raise exc



from os import PathLike
from typing import List, Optional, Union

from ..dataset import Dataset
from ..geometry import BoundingBox, Mag
from ..utils import warn_deprecated
from ._download_dataset import download_dataset as actual_download_dataset


def download_dataset(
    dataset_name: str,
    organization_name: str,
    bbox: Optional[BoundingBox] = None,
    layers: Optional[List[str]] = None,
    mags: Optional[List[Mag]] = None,
    path: Optional[Union[PathLike, str]] = None,
    exist_ok: bool = False,
) -> Dataset:
    warn_deprecated("download_dataset", "Dataset.download")
    return actual_download_dataset(
        dataset_name,
        organization_id=organization_name,
        bbox=bbox,
        layers=layers,
        mags=mags,
        path=path,
        exist_ok=exist_ok,
    )



from typing import List, NamedTuple, Optional
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from ..geometry import BoundingBox
from ..geometry.bounding_box import _DEFAULT_BBOX_NAME
from .utils import Vector3, enforce_not_null, filter_none_values

DEFAULT_BOUNDING_BOX_COLOR = [0.2, 0.5, 0.1, 1]


class Parameters(NamedTuple):
    name: str  # dataset name
    scale: Vector3  # dataset voxel_size
    description: Optional[str] = None
    organization: Optional[str] = None
    offset: Optional[Vector3] = None  # deprecated. Kept for backward compatibility.
    time: Optional[
        int
    ] = None  # UNIX timestamp marking the creation time & date of an annotation.
    editPosition: Optional[Vector3] = None
    editRotation: Optional[Vector3] = None
    zoomLevel: Optional[float] = None
    taskBoundingBox: Optional[BoundingBox] = None
    userBoundingBoxes: Optional[List[BoundingBox]] = None

    def _dump_bounding_box(
        self,
        xf: XmlWriter,
        bounding_box: BoundingBox,
        tag_name: str,
        bbox_id: Optional[int],  # user bounding boxes need an id
    ) -> None:
        color = bounding_box.color or DEFAULT_BOUNDING_BOX_COLOR

        attributes = {
            "name": _DEFAULT_BBOX_NAME
            if bounding_box.name is None
            else str(bounding_box.name),
            "isVisible": "true" if bounding_box.is_visible else "false",
            "color.r": str(color[0]),
            "color.g": str(color[1]),
            "color.b": str(color[2]),
            "color.a": str(color[3]),
            "topLeftX": str(bounding_box.topleft.x),
            "topLeftY": str(bounding_box.topleft.y),
            "topLeftZ": str(bounding_box.topleft.z),
            "width": str(bounding_box.size.x),
            "height": str(bounding_box.size.y),
            "depth": str(bounding_box.size.z),
        }
        if bbox_id is not None:
            attributes["id"] = str(bbox_id)

        xf.tag(
            tag_name,
            attributes,
        )

    def _dump_task_bounding_box(self, xf: XmlWriter) -> None:
        task_bounding_box = getattr(self, "taskBoundingBox")
        if task_bounding_box is not None:
            self._dump_bounding_box(xf, task_bounding_box, "taskBoundingBox", None)

    def _dump_user_bounding_boxes(self, xf: XmlWriter) -> None:
        user_bounding_boxes = getattr(self, "userBoundingBoxes")

        if user_bounding_boxes is not None:
            # pylint: disable=not-an-iterable

            # User bounding boxes need an id to be recognized
            # when uploaded to webknossos, which is added by bbox_idx:
            for bbox_idx, user_bounding_box in enumerate(
                user_bounding_boxes,
            ):
                self._dump_bounding_box(
                    xf, user_bounding_box, "userBoundingBox", bbox_idx
                )

    def _dump(self, xf: XmlWriter) -> None:
        xf.startTag("parameters")
        xf.tag(
            "experiment",
            filter_none_values(
                {
                    "name": self.name,
                    "organization": self.organization,
                    "description": self.description,
                }
            ),
        )
        xf.tag(
            "scale",
            {
                "x": str(float(self.scale[0])),
                "y": str(float(self.scale[1])),
                "z": str(float(self.scale[2])),
            },
        )

        if self.offset is not None:
            # pylint: disable=unsubscriptable-object
            xf.tag(
                "offset",
                {
                    "x": str(float(self.offset[0])),
                    "y": str(float(self.offset[1])),
                    "z": str(float(self.offset[2])),
                },
            )

        if self.time is not None:
            xf.tag("time", {"ms": str(self.time)})
        if self.editPosition is not None:
            # pylint: disable=unsubscriptable-object
            xf.tag(
                "editPosition",
                {
                    "x": str(float(self.editPosition[0])),
                    "y": str(float(self.editPosition[1])),
                    "z": str(float(self.editPosition[2])),
                },
            )
        if self.editRotation is not None:
            # pylint: disable=unsubscriptable-object
            xf.tag(
                "editRotation",
                {
                    "xRot": str(float(self.editRotation[0])),
                    "yRot": str(float(self.editRotation[1])),
                    "zRot": str(float(self.editRotation[2])),
                },
            )
        if self.zoomLevel is not None:
            xf.tag("zoomLevel", {"zoom": str(self.zoomLevel)})

        self._dump_task_bounding_box(xf)
        self._dump_user_bounding_boxes(xf)

        xf.endTag()  # parameters

    @classmethod
    def _parse_bounding_box(cls, bounding_box_element: Element) -> BoundingBox:
        topleft = (
            int(bounding_box_element.get("topLeftX", 0)),
            int(bounding_box_element.get("topLeftY", 0)),
            int(bounding_box_element.get("topLeftZ", 0)),
        )
        size = (
            int(bounding_box_element.get("width", 0)),
            int(bounding_box_element.get("height", 0)),
            int(bounding_box_element.get("depth", 0)),
        )
        color = None
        if bounding_box_element.get("color.r"):  # also checks for empty strings
            color = (
                float(enforce_not_null(bounding_box_element.get("color.r"))),
                float(enforce_not_null(bounding_box_element.get("color.g"))),
                float(enforce_not_null(bounding_box_element.get("color.b"))),
                float(enforce_not_null(bounding_box_element.get("color.a"))),
            )

        return BoundingBox(
            topleft,
            size,
            name=bounding_box_element.get("name"),
            is_visible=bounding_box_element.get("isVisible", "true") == "true",
            color=color,
        )

    @classmethod
    def _parse_user_bounding_boxes(cls, nml_parameters: Element) -> List[BoundingBox]:
        if nml_parameters.find("userBoundingBox") is None:
            return []
        bb_elements = nml_parameters.findall("userBoundingBox")
        return [cls._parse_bounding_box(bb_element) for bb_element in bb_elements]

    @classmethod
    def _parse_task_bounding_box(cls, nml_parameters: Element) -> Optional[BoundingBox]:
        bb_element = nml_parameters.find("taskBoundingBox")
        if bb_element is not None:
            return cls._parse_bounding_box(bb_element)

        return None

    @classmethod
    def _parse(cls, nml_parameters: Element) -> "Parameters":
        offset = None
        if nml_parameters.find("offset") is not None:
            offset_element = enforce_not_null(nml_parameters.find("offset"))
            offset = (
                float(offset_element.get("x", 0)),
                float(offset_element.get("y", 0)),
                float(offset_element.get("z", 0)),
            )

        editRotation = None
        if nml_parameters.find("editRotation") is not None:
            edit_rotation_element = enforce_not_null(
                nml_parameters.find("editRotation")
            )
            editRotation = (
                float(edit_rotation_element.get("xRot", 0)),
                float(edit_rotation_element.get("yRot", 0)),
                float(edit_rotation_element.get("zRot", 0)),
            )

        editPosition = None
        if nml_parameters.find("editPosition") is not None:
            editPosition = (
                float(
                    enforce_not_null(nml_parameters.find("editPosition")).get("x", 0)
                ),
                float(
                    enforce_not_null(nml_parameters.find("editPosition")).get("y", 0)
                ),
                float(
                    enforce_not_null(nml_parameters.find("editPosition")).get("z", 0)
                ),
            )

        time = None
        if nml_parameters.find("time") is not None:
            time = int(enforce_not_null(nml_parameters.find("time")).get("ms", 0))

        zoomLevel: Optional[float] = None
        if nml_parameters.find("zoomLevel") is not None:
            zoom_str = enforce_not_null(nml_parameters.find("zoomLevel")).get("zoom", 0)
            try:
                zoomLevel = int(zoom_str)
            except ValueError:
                zoomLevel = float(zoom_str)

        taskBoundingBox = cls._parse_task_bounding_box(nml_parameters)
        userBoundingBoxes = cls._parse_user_bounding_boxes(nml_parameters)

        scale_element = enforce_not_null(nml_parameters.find("scale"))
        experiment_element = enforce_not_null(nml_parameters.find("experiment"))
        return cls(
            name=experiment_element.get("name", "Unnamed Experiment"),
            description=experiment_element.get("description"),
            organization=experiment_element.get("organization"),
            scale=(
                float(scale_element.get("x", 0)),
                float(scale_element.get("y", 0)),
                float(scale_element.get("z", 0)),
            ),
            offset=offset,
            time=time,
            editPosition=editPosition,
            editRotation=editRotation,
            zoomLevel=zoomLevel,
            taskBoundingBox=taskBoundingBox,
            userBoundingBoxes=userBoundingBoxes,
        )



from pathlib import Path


def guess_if_segmentation_path(filepath: Path) -> bool:
    lowercase_filepath = str(filepath).lower()
    return any(i in lowercase_filepath for i in ["segmentation", "labels"])
